/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// Implementation of CGif

// Much of this code was adapted from xgif.c by John Bradley,
// University of Pennsylvania (bradley@cis.upenn.edu)

#define DEBUG_GIF 0
#define DEBUG_CONTROL 0
#define DEBUG_MERGE 0
#define DEBUG_SNPRINTF 0
#define DEBUG_EXC 0

#ifdef WIN32
# if DEBUG_EXC
#  include <windows.h>
#  include <malloc.h>
# endif
# if DEBUG_SNPRINTF
#  define snprintf _snprintf
# endif
#endif

#include <limits.h>

#include <Xm/List.h>
#include <Xm/Text.h>

#include "defaultColors.h"

#include "CSGA.h"

// KE: The following is the prototype for _XGetBitsPerPixel, which is
// a convenience function found in the MIT distribution in
// /opt/local/share/src/R5/mit/lib/X/XImUtil.c (at the APS)

extern "C" {
int _XGetBitsPerPixel(Display *dpy, int depth);
}

// Debug utilities

#if DEBUG_GIF
void CGif::print1(const char *name) const
{
    print("%s: ptr=%x pos=%d fileSize=%d\n",
      name,ptr,(int)(ptr-raw),fileSize);
}
#endif

#if DEBUG_EXC
void checkHeap(char *title)
{
  // Print the title
    if(title) print("%s",title);
    
  // Check heap status
    int heapStatus = _heapchk();
    switch(heapStatus ) {
    case _HEAPOK:
	print(" Heap is OK\n");
	break;
    case _HEAPEMPTY:
	print(" Heap is empty\n");
	break;
    case _HEAPBADBEGIN:
	print(" Heap Error: Bad start\n");
	break;
    case _HEAPBADNODE:
	print(" Heap Error: Bad node\n");
	break;
    default:
	print(" Heap Error: Unknown return value [%d]\n",
	  heapStatus);
	break;
    }
}

void dumpHeap(char *title)
{
    _HEAPINFO hinfo;
    int heapStatus;
    hinfo._pentry = NULL;
  // Print the title
    if(title) print("%s\n",title);
    
  // Walk the heap
    while((heapStatus = _heapwalk(&hinfo)) == _HEAPOK ) {
	print("%6s block at %Fp of size %4.4X\n",
	  (hinfo._useflag == _USEDENTRY ? "USED" : "FREE" ),
          hinfo._pentry, hinfo._size );
    }
    
    switch(heapStatus ) {
    case _HEAPOK:
     	print("Dump Heap: Heap is OK\n");
	break;
    case _HEAPEMPTY:
	print("Dump Heap: Heap is empty\n");
	break;
    case _HEAPBADBEGIN:
	print(" Dump Heap: Bad start\n");
	break;
    case _HEAPBADNODE:
	print(" Dump Heap: Bad node\n");
	break;
    case _HEAPBADPTR:
	print(" Dump Heap: Bad pointer to heap\n");
	break;
    case _HEAPEND:
	print(" Dump Heap: End of Heap\n");
	break;
    default:
	print(" Dump Heap: Unknown return value [%d]\n",
	  heapStatus);
	break;
    }
}
#endif

CGif::CGif(int nColorsIn) :
  // Note these are initialized in declaration order, not this order
  // Be sure this list is the same for both constructors
    control(NULL),
    state(ST_ALLDONE),
    dataMode(GRAPHIC_MODE),
    bgMode(BACKGROUND_GIF),
    edited(False),
    nImages(0),
    raw(NULL),
    ptr(NULL),
    eof(NULL),
    ScreenDepth(DefaultDepth(display,screen)),
    XBitsPerPixel(_XGetBitsPerPixel(display,ScreenDepth)),
    XBitmapBitOrder(BitmapBitOrder(display)),
    XBitmapPad(BitmapPad(display)),
    XBitmapUnit(BitmapUnit(display)),
    XByteOrder(ImageByteOrder(display)),
    BytesPerPixel(XBitsPerPixel/8),
    visual(DefaultVisual(display,screen)),
    ClientByteOrder(MSBFirst)
{
    int *def=NULL;

  // Fill in private data that would be obtained from a file
    *fileName='\0';
    fileSize=0;
    strcpy(GifType,"GIF87a");
    LogicalScreenWidth=0;
    LogicalScreenHeight=0;
    GlobalColorTableFlag=True;
    GlobalColorTableSortFlag=False;
    ColorResolution=ScreenDepth-1;
    BackgroundColorIndex=0;
    PixelAspectRatio=0;

    setNColors(nColorsIn);

  // Make color table
    switch(nColors) {
    case 2:
	def=defaultColors2;
	break;
    case 4:
	def=defaultColors4;
	break;
    case 8:
	def=defaultColors8;
	break;
    case 16:
	def=defaultColors16;
	break;
    case 32:
	def=defaultColors32;
	break;
    case 64:
	def=defaultColors64;
	break;
    case 128:
	def=defaultColors128;
	break;
    case 256:
	def=defaultColors256;
	break;
    }

  // Fill in the Red, Green, and Blue arrays
    for(int i=0; i < nColors; i++) {
	Red[i]=def[3*i];
	Green[i]=def[3*i+1];
	Blue[i]=def[3*i+2];
    }
    
  // Allocate the colors
    allocateColors();

  // Allocate a Global block
    CGlobal *global=new CGlobal(this);
    if(!global) {
	errMsg("CGif: Cannot allocate Global");
	return;
    }
    blockList.add(*(CBlock *)global);
}


CGif::CGif() :
  // Note these are initialized in declaration order, not this order
  // Be sure this list is the same for both constructors
    control(NULL),
    state(ST_ALLDONE),
    dataMode(GRAPHIC_MODE),
    bgMode(BACKGROUND_GIF),
    edited(False),
    nImages(0),
    raw(NULL),
    ptr(NULL),
    eof(NULL),
    ScreenDepth(DefaultDepth(display,screen)),
    XBitsPerPixel(_XGetBitsPerPixel(display,ScreenDepth)),
    XBitmapBitOrder(BitmapBitOrder(display)),
    XBitmapPad(BitmapPad(display)),
    XBitmapUnit(BitmapUnit(display)),
    XByteOrder(ImageByteOrder(display)),
    BytesPerPixel(XBitsPerPixel/8),
    visual(DefaultVisual(display,screen)),
    ClientByteOrder(MSBFirst)
{
}

CGif::~CGif(void)
{
    cleanup();
}

// Loads the file
int CGif::load(char *fileNameIn)
{
    int status;

#if DEBUG_EXC
    print("load: Before cleanup\n");
#endif
  // Clean up any old data */
    cleanup();

  // Store the fileName
#if DEBUG_EXC
    print("load: Before strncpy\n");
#endif
    strncpy(fileName,fileNameIn,PATH_MAX);
    fileName[PATH_MAX-1]='\0';

  // Open the file
#if DEBUG_EXC
    print("load: Before openFile\n");
#endif
    status=openFile(fileName);
    if(status != SGA_OK) return status;

  // Parse the header
#if DEBUG_EXC
    print("load: Before parseHeader\n");
#endif
    status=parseHeader();
    if(status != SGA_OK) return status;

  // Parse the global color table
#if DEBUG_EXC
    print("load: Before parseGlobalColorTable\n");
#endif
    status=parseGlobalColorTable();
    if(status != SGA_OK) return status;

  // Allocate colors
#if DEBUG_EXC
    print("load: Before allocateColors\n");
#endif
    status=allocateColors();
    if(status != SGA_OK) return status;

  // Parse the rest of the file
#if DEBUG_EXC
    print("load: Before parseRemainderOfFile\n");
#endif
    status=parseRemainderOfFile();
    if(status != SGA_OK) return status;

#if DEBUG_EXC
    print("load: At end\n");
#endif
    return SGA_OK;
}

// Writes the file
int CGif::writeFile(const char *fileNameIn)
{
    FILE *fp=NULL;
    int status,index,width,height;
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;

#if DEBUG_GIF
    print("CGif::writeFile");
#endif

  // Try to open the file
    fp=fopen(fileNameIn,"wb");
    if(fp == NULL) {
	errMsg("CGif::writeFile: Cannot open file:\n"
	  "  %s\n",fileNameIn);
	goto CLEANUP;
    }

  // Insure the logical extents are correct
    width=LogicalScreenWidth;
    height=LogicalScreenHeight;
    setLogicalScreenWidth(width);
    setLogicalScreenHeight(height);
    if(width != LogicalScreenWidth || height != LogicalScreenHeight) {
	errMsg("CGf::writeFile: Changed logical screen width and height to "
	  "hold largest image\n"
	  "  Logical Screen Width:  %d -> %d\n"
	  "  Logical Screen Height: %d -> %d\n",
	  width,LogicalScreenWidth,height,LogicalScreenHeight);
	if(selectedBlock  && selectedBlock->getType() == TYPE_GLOBAL) {
	    selectedBlock->displayData();
	}
    }

  // Let each block write itself
    index=1;
    while(iter != eol) {
	CBlock *block=iter;
	status=block->write(fp);
	if(status != SGA_OK) {
	    errMsg("CGf::writeFile: Could not write %s block %d",
	      block->stringType(),index);
	    goto CLEANUP;
	}
	++index;
	++iter;
    }

  // Write the trailer
    fputc(TRAILER,fp);

  // Close the file
    fclose(fp);
#if 0
    print("File %s has been written\n",fileNameIn);
#endif
    return SGA_OK;

  // Clean up
  CLEANUP:
    if(fp) fclose(fp);
    return SGA_ERROR;
}

// Cleans up after errors reading file
void CGif::cleanup(void)
{
#if DEBUG_EXC
    print("cleanup: Before delete raw\n");
    checkHeap(" Check Heap:");
  //    dumpHeap(" Dump Heap:");
#endif
  // Free the raw data
    if(raw) {
	delete [] raw;
	raw=NULL;
    }
#if DEBUG_EXC
    print("cleanup: Before clean blockList\n");
    checkHeap(" Check Heap:");
  //    dumpHeap(" Dump Heap:");
#endif
  // Clear the blockList and delete the blocks
    while(getBlockCount() > 0) {
	CBlock *pLast=blockList.last();
	blockList.remove(*pLast);
#if DEBUG_EXC
	print("cleanup: Before delete pLast=%x [%s]\n",
	  pLast,pLast->stringType());
	checkHeap(" Check Heap:");
      //	dumpHeap(" Dump Heap:");
#endif
	delete pLast;
    }
}

// Returns a pointer to the short part of the filename
const char *CGif::getShortName(void) const
{
    const char *p;

  // Get the short name
    p=strrchr(fileName,DIR_DELIMITER_CHAR);
    if(p) {
	return p+1;
    } else {
	return fileName;
    }
}

// Sets the filename
void CGif::setLongName(const char *fileNameIn)
{
    strncpy(fileName,fileNameIn,PATH_MAX);
    fileName[PATH_MAX-1]='\0';
}

#define NAME_LENGTH 26
void CGif::printSummary(void) const
{
    static char header[]=
  //   12345678911234567892123456789312345678941234567895123456789612345678971234567898
      "Name                      Ver    W    H Col Frm  Bg Trn Inl Off App Com Txt";
    char line[100];
    int nCurrentImages=0,trn=0,inl=0,off=0,app=0,com=0,txt=0;
    char name[NAME_LENGTH];

  // Get the short name
    strncpy(name,getShortName(),NAME_LENGTH);
    name[NAME_LENGTH-1] = '\0';

  // Check which attributes occur
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_APPLICATION) app++;
	if(iter->getType() == TYPE_COMMENT) com++;
	if(iter->getType() == TYPE_PLAINTEXT) txt++;
	if(iter->getType() == TYPE_IMAGE) {
	    nCurrentImages++;
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    if(image->getTransparentColorFlag()) trn++;
	    if(image->getInterlace()) inl++;
	    if(image->getTopOffset() || image->getLeftOffset()) off++;
	}
	++iter;
    }

  // Print the results
    sprintf(line,"\n%s\n",header);
    print(line);
    sprintf(line,"%-25s %3s %4d %4d %3d %3d %3d %3d %3d %3d %3d %3d %3d\n",
      name,GifType+3,LogicalScreenWidth,LogicalScreenHeight,nColors,
      nCurrentImages,BackgroundColorIndex,
      trn,inl,off,app,com,txt);
    print(line);
}

// Opens and reads the file, storing the information as rawData
int CGif::openFile(char *fname)
{
    FILE *fp;

#if DEBUG_GIF
    print("CGif::openFile");
#endif

#if DEBUG_SNPRINTF
  // vsnprintf test
    char buf[10]=".........";
    for(int i=0; i < 16; i++) {
	int ret=snprintf(buf,i,"012345");
	buf[9]='\0';
	print("%2d ret=%d buf=%s\n",i,ret,buf);
    }
#endif

#if DEBUG_EXC
    print("openFile: Before fopen\n");
#endif
  // Try to open the file
    state=ST_START;
    ptr=eof=NULL;
    fileSize=0;
#ifdef WIN32
  // WIN32 opens files in text mode by default and then throws out CRLF
    fp=fopen(fname,"rb");
#else
    fp=fopen(fname,"r");
#endif
    if(fp == NULL) {
	errMsg("CGif::openFile: Cannot open file:\n"
	  "  %s\n",fname);
	goto CLEANUP;
    }

  // Find the size of the file
    fseek(fp,0L,SEEK_END);
    fileSize=(int)ftell(fp);
    fseek(fp,0L,SEEK_SET);

#if DEBUG_EXC
    print("openFile: Before allocating raw\n");
    print(" fileSize=%d\n",fileSize);
#ifdef WIN32
    {
	HANDLE hHeap=GetProcessHeap();
	BOOL res=HeapValidate(hHeap,0,NULL);
	print(" Process Heap 0x%x is %s\n",hHeap,res?"Valid":"Invalid");
	checkHeap(" Heap Check:");
    }
#endif
#if 0
    try {
	raw=new unsigned char[fileSize];
    } catch (...) {
	print("openFile: Error allocating raw\n");
    }
    print("openFile: After raw=new Byte[fileSize];\n");
    if(raw) delete [] raw;
    print("openFile: After if(raw) delete [] raw;\n");
#endif
#endif
  // Allocate memory for the raw data
    if(!(raw=new Byte[fileSize])) {
#if DEBUG_EXC
	print("openFile: Before errMsg\n");
#endif
	errMsg("loadGIF: Not enough memory for raw data:\n"
	  "  %s\n",fname);
	goto CLEANUP;
    }
    
#if DEBUG_EXC
    print("openFile: Before fread\n");
#ifdef WIN32
    {
	HANDLE hHeap=GetProcessHeap();
	BOOL res=HeapValidate(hHeap,0,NULL);
	print(" Process Heap 0x%x is %s\n",hHeap,res?"Valid":"Invalid");
	checkHeap(" Heap Check:");
    }
#endif
#endif
  // Read the file in one chunk
    if(fread((char *)raw,fileSize,1,fp) != 1) {
	char *errstring=strerror(ferror(fp));

	errMsg("CGif::openFile: Cannot read file:\n"
	  "  %s\n  %s\n",fname,errstring);
	goto CLEANUP;
    }
    fclose(fp);
    eof=raw+fileSize;
    state=ST_FILEREAD;
#if DEBUG_EXC
    print("openFile: Before return SGA_OK\n");
#endif
    return SGA_OK;

  // Clean up
  CLEANUP:
#if DEBUG_EXC
    print("openFile: CLEANUP: Before fclose\n");
#endif
    if(fp) fclose(fp);
    cleanup();
#if DEBUG_EXC
    print("openFile: Before return SGA_ERROR\n");
#endif
    return SGA_ERROR;
}

// Parses the header
int CGif::parseHeader(void)
{
    Byte ch;    

#if DEBUG_GIF
    print1("CGif::parseHeader");
#endif

  // Allocate a Global block
    CGlobal *global=new CGlobal(this);
    if(!global) {
	errMsg("parseHeader: Cannot allocate Global for %s\n",
	  fileName);
	return SGA_ERROR;
    }
    blockList.add(*(CBlock *)global);
    
  // Initialize the pointer to the start of the raw data
    ptr=raw;

  // We expect 13 bytes
    if(!checkRemainingSize(13)) {
	errMsg("CGif::parseHeader: File ends prematurely\n  %s\n",
	  fileName);
	cleanup();
	return SGA_ERROR;
    }

  // Parse the Signature (3 bytes) and the Version (3 bytes)
    if(!strncmp((char *)ptr, "GIF89a", 6)) {
    } else if(!strncmp((char *)ptr, "GIF87a", 6)) {
    } else {
	errMsg("CGif::parseHeader: Not a valid GIF file\n  %s\n",
	  fileName);
	cleanup();
	return SGA_ERROR;
    }
    strncpy(GifType,(char *)ptr,6);
    GifType[6]='\0';
    ptr += 6;

  // Parse the Logical Screen Descriptor
  // Parse the required ScreenWidth
    ch=NEXTBYTE;
    LogicalScreenWidth=ch+0x100*NEXTBYTE;

  // Parse the required ScreenHeight
    ch=NEXTBYTE;
    LogicalScreenHeight=ch+0x100*NEXTBYTE;

#if DEBUG_GIF
    {
	static int ifirst=1;
	
	if(ifirst) {
	    int depth=DefaultDepth(display,screen);
	    int bitmap_bit_order=BitmapBitOrder(display);
	    int bitmap_pad=BitmapPad(display);
	    int bitmap_unit=BitmapUnit(display);
	    int byte_order=ImageByteOrder(display);
	    int bits_per_pixel=_XGetBitsPerPixel(display,depth);
	    
	    ifirst=0;
	    print("\nX Server Image Properties:\n");
	    print("  depth=%d [%d colors (0-%x)]\n",depth,1<<depth,
	      (int)(1<<depth)-1);
	    print("  bitmap_bit_order=%d [LSBFirst=%d MSBFirst=%d]\n",
	      bitmap_bit_order,LSBFirst,MSBFirst);
	    print("  bitmap_pad=%d\n",bitmap_pad);
	    print("  bitmap_unit=%d\n",bitmap_unit);
	    print("  byte_order=%d [LSBFirst=%d MSBFirst=%d]\n",
	      byte_order,LSBFirst,MSBFirst);
	    print("  bits_per_pixel=%d\n",bits_per_pixel);
	}
    }
#endif    

  // Parse the Packed byte
  // Bits 0-2: MASK210 Size of the Global Color Table
  // Bit 3:            Color Table Sort Flag (Ignored)
  // Bits 4-6:         Color Resolution (Ignored)
  // Bit 7:    MASK7   Global Color Table Flag
    ch=NEXTBYTE;
    GlobalColorTableFlag=((ch & MASK7) ? True : False);
    GlobalColorTableSortFlag=((ch & MASK3) ? True : False);
    SizeOfGlobalColorTable=(ch & MASK210);
    ColorResolution=(ch & MASK654)>>4;
    BitsPerPixel=SizeOfGlobalColorTable+1;
    ColorTableEntries=1<<BitsPerPixel;
    BitMask=ColorTableEntries-1;

  // Parse the BackgroundColorIndexColor
    BackgroundColorIndex=NEXTBYTE;

  // Parse the PixelAspectRatio, not used.
    PixelAspectRatio=NEXTBYTE;

  // End of header
#if DEBUG_GIF
    char version[7];
    
    strncpy(version,(char *)raw,6);
    version[6]='\0';
    print("\nloadGIF: %s fileSize=%d\n",fileName,fileSize);
    print("HEADER Version=%s\n",version);
    print("  LogicalScreenWidth=%d LogicalScreenHeight=%d\n",
      LogicalScreenWidth,LogicalScreenHeight);
    print("  GlobalColorTableFlag=%s Sorted=%s\n",
      GlobalColorTableFlag?"True":"False",
      GlobalColorTableSortFlag?"True":"False");
    print("  SizeOfGlobalColorTable=%d (%d-bit, %d colors)\n",
      SizeOfGlobalColorTable,
      BitsPerPixel,ColorTableEntries);
    print("  BackgroundColorIndex=%d PixelAspectRatio=%d\n",
      BackgroundColorIndex,PixelAspectRatio);
#endif

    state=ST_HEADERDONE;
    return SGA_OK;
}

// Parses the color table
int CGif::parseGlobalColorTable(void)
{
    int i;

#if DEBUG_GIF
    print1("CGif::parseGlobalColorTable");
#endif

  // Parse the Global Color Table if present
  // KE: Note that if there is no global table, there may be local tables
    if(GlobalColorTableFlag) {
      // We expect 3*ColorTableEntries bytes
	if(!checkRemainingSize(3*ColorTableEntries)) {
	    errMsg("CGif::parseColorTable: File ends prematurely\n  %s\n",
	      fileName);
	    cleanup();
	    return SGA_ERROR;
	}
	nColors=ColorTableEntries;
	for (i=0; i < ColorTableEntries; i++) {
	    Red[i]=NEXTBYTE;
	    Green[i]=NEXTBYTE;
	    Blue[i]=NEXTBYTE;
#if DEBUG_GIF > 1
	    print("%3d %2x %2x %2x\n",i,Red[i],Green[i],Blue[i]);
#endif	    
	}
    } else {
      // No global colormap in GIF file
        errMsg("parseColorTable:  No global color table in %s\n",
	  fileName);
        nColors=0;
    }
    state=ST_COLORDONE;

    return SGA_OK;
}

// Allocates the X colors
int CGif::allocateColors(void)
{
    Pixel blackPixel=BlackPixel(display,screen);
    Colormap colormap=DefaultColormap(display,screen);
    XColor xColors[256];
    int unallocated=0;

    for(int i=0; i < nColors; i++) {
	xColors[i].red=Red[i]<<8;
	xColors[i].green=Green[i]<<8;
	xColors[i].blue=Blue[i]<<8;
	xColors[i].flags=DoRed|DoGreen|DoBlue;
	if(!XAllocColor(display,colormap,&xColors[i])) { 
	    unallocated++;
	    xColors[i].pixel=blackPixel;
	}
	pixels[i]=xColors[i].pixel;
    }
    if(unallocated) {
	errMsg("CGif: allocateColors: Could not allocate %d of %d colors",
	  unallocated,nColors);
    }
    if(BackgroundColorIndex >= 0 && BackgroundColorIndex < nColors) {
	BackgroundColor=pixels[BackgroundColorIndex];
    } else {
	BackgroundColor=WhitePixel(display,screen);
    }

    return SGA_OK;
}

// Parses the rest of the file after the color table
int CGif::parseRemainderOfFile(void)
{
    Boolean error, done;
    register Byte ch;
    
#if DEBUG_GIF
    print("CGif::parseRemainderOfFile");
#endif

  // Parse the data until the terminator is reached
    error=True;
    done=False;
    while(!done && ptr < eof) {
	ch=NEXTBYTE;
	switch(ch) {
	case IMAGESEP:
	{
#if DEBUG_GIF
	    print("IMAGE [%2x]\n",ch);
#endif
	  // If there is not a Control, make a default one
	    if(!control) {
		control=new CControl(this);
		if(!control) {
		    errMsg("CGif::parseRemainderOfFile: "
		      "Cannot allocate Control for %s\n",
		      fileName);
		    return SGA_ERROR;
		}
#if DEBUG_CONTROL
		print("CGif::parseRemainderOfFile: Created  Control %x\n",
		  control);
#endif
	    }
	  // Make an Image and add it to the block list
	    CImage *image=new CImage(this);
	    if(!image) {
		errMsg("parseRemainderOfFile: "
		  "Cannot allocate Image for %s\n",
		  fileName);
		return SGA_ERROR;
	    }
	    blockList.add(*(CBlock *)image);
	    image->parse();
	    break;
	}
	case EXTENSION:
#if DEBUG_GIF
	    print("EXTENSION [%2x]\n",ch);
#endif	    
	    if(!parseExtension()) done=True;
	    break;
	case TRAILER:
#if DEBUG_GIF
	    print("TRAILER [%2x]\n",ch);
#endif	    
	    done=True;
	    error=False;
	    break;
	default:
#if DEBUG_GIF
	    print("UNKNOWN [%2x]\n",ch);
#endif	    
	    errMsg("CGif::parseRemainderOfFile: "
	      "Unknown block type [%02x] for %s\n",
	      ch,fileName);
	    done=True;
	    return SGA_ERROR;
	}
    }

  // Free the raw data
    if(raw) {
	delete [] raw;
	raw=NULL;
    }
  // Free any left-over Control
    if(control) {
	delete control;
	control=NULL;
    }

    if(!done) {
	errMsg("parseRemainderOfFile: Terminator not found for %s\n",
	  fileName);
    }
    state=ST_ALLDONE;
    return SGA_OK;
}

int CGif::parseExtension(void)
{
    Byte ch;
    
#if DEBUG_GIF
    print1("CGif::parseExtension");
#endif

    ch=NEXTBYTE;
    switch(ch) {
    case GRAPHCONTROL:     // Should be 0xF9 (249)
    {
#if DEBUG_GIF
	print("  GRAPHCONTROL [%2x]\n",ch);
#endif
      // The CControl is private data and is not in the blockList
      // It will be moved to the Image and the Image will delete it
      // Delete any existing control, only the last one is important
	if(control) {
#if DEBUG_CONTROL
	    print("CGif::parseExtension: Created  Control %x\n",
	      control);
#endif
	    delete control;
	}
	control=new CControl(this);
	if(!control) {
	    errMsg("parseExtension: Cannot allocate Control for %s\n",
	      fileName);
	    return SGA_ERROR;
	}
#if DEBUG_CONTROL
		print("CGif::parseExtension: Created  Control %x\n",
		  control);
#endif
      // (Not added to the blocklist)
	control->parse();
	break;
    }
    case PLAINTEXT:     // Should be 0x01 (1)
    {
#if DEBUG_GIF
	print("  PLAINTEXT [%2x]\n",ch);
#endif	    
	CPlaintext *plaintext=new CPlaintext(this);
	if(!plaintext) {
	    errMsg("CGif::parseExtension: Cannot allocate Plaintext for %s\n",
	      fileName);
	    return SGA_ERROR;
	}
	blockList.add(*(CBlock *)plaintext);
	plaintext->parse();
	break;
    }
    case APPLICATION:     // Should be 0x0B (11)
    {
#if DEBUG_GIF
	print("  APPLICATION [%2x]\n",ch);
#endif	    
	CApplication *application=new CApplication(this);
	if(!application) {
	    errMsg("CGif::parseExtension: Cannot allocate Application for %s\n",
	      fileName);
	    return SGA_ERROR;
	}
	blockList.add(*(CBlock *)application);
	application->parse();
	break;
    }
    case COMMENT:     // Should be 0xFE (254)
    {
#if DEBUG_GIF
	print("  COMMENT [%2x]\n",ch);
#endif	    
	CComment *comment=new CComment(this);
	if(!comment) {
	    errMsg("CGif::parseExtension: Cannot allocate Comment for %s\n",
	      fileName);
	    return SGA_ERROR;
	}
	blockList.add(*(CBlock *)comment);
	comment->parse();
	break;
    }
    default:
#if DEBUG_GIF
	print("  UNKNOWN [%2x]\n",ch);
#endif	    
	errMsg("CGif::parseExtension: Unknown extension [%02x] for %s\n",
	  ch,fileName);
	return SGA_ERROR;
    }
    return SGA_ERROR;
}

int CGif::updateListArea(void)
{
    XmString xmString;

    XmListDeleteAllItems(listArea);
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    char string[100];

	    sprintf(string,"%s %d",iter->stringType(),image->getImageNumber());
	    xmString=XmStringCreateLtoR(string,XmSTRING_DEFAULT_CHARSET);
	} else {
	    xmString=XmStringCreateLtoR(iter->stringType(),
	      XmSTRING_DEFAULT_CHARSET);
	}
	XmListAddItemUnselected(listArea,xmString,0);
	XmStringFree(xmString);
	++iter;
    }

    return SGA_OK;
}


// Prints out the items in the block list.  Intended for debugging.
void CGif::dumpBlockList(void) const
{
    int pos=1;

    print("Dumping blockList for %s  with count=%d\n",
      fileName?fileName:"New Block",blockList.count());
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	CBlock *block=(CBlock *)iter;
	print("  Block %d: Type=%-13s Address=%x\n",
	  pos,iter->stringType(),block);
	++iter;
	++pos;
    }
}

// Returns a pointer to the block at position pos.  If remove is true,
// removes it from the list.
CBlock *CGif::getBlock(const int pos, Boolean remove)
{
    int i=1;     // First position is 1

  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
#if DEBUG_MERGE > 1
	CBlock *b=iter;
	print("CGif::getBlock: i=%d pos=%d block=%x count=%d remove=%d(%s)\n",
	  i,pos,b,blockList.count(),remove,remove?"True":"False");
#endif
	if(i == pos) {
	    CBlock *block=iter;
	    if(remove) {
		blockList.remove(*block);
#if DEBUG_MERGE > 1
		print("  removed, count is now: %d\n",blockList.count());
#endif
	    }
	    return block;
	}
	++i;
	++iter;
    }
    return NULL;
}

int CGif::getPos(const CBlock *block) const
{
    int i=1;     // First position is 1

  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter == block) {
	    return i;
	}
	++i;
	++iter;
    }
    return 0;
}

CBlock *CGif::getFirstImage(void) const
{
  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    return iter;
	}
	++iter;
    }
    return NULL;
}

CBlock *CGif::getLastImage(void) const
{
  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.last());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    return iter;
	}
	--iter;
    }
    return NULL;
}

CBlock *CGif::getNextImage(const CBlock *block) const
{
    int found=0;

  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(!found) {
	    if(iter == block) found=1;
	} else if(iter->getType() == TYPE_IMAGE) {
	    return iter;
	}
	++iter;
    }
    return NULL;
}

CBlock *CGif::getPrevImage(const CBlock *block) const
{
    int found=0;

  // Find the block by parsing the list (inefficient)
    tsDLIterBD<CBlock> iter(blockList.last());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(!found) {
	    if(iter == block) found=1;
	} else if(iter->getType() == TYPE_IMAGE) {
	    return iter;
	}
	--iter;
    }
    return NULL;
}

CBlock *CGif::deleteBlock(CBlock *block)
{
    int found=0;

  // Check for the header
    if(block->getType() == TYPE_GLOBAL) {
	errMsg("CGif::deleteBlock: Cannot delete the Global block");
	return NULL;
    }

  // Find the block by parsing the list (inefficient)
    CBlock *tmp=blockList.first();;
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if((CBlock *)iter == block) {
	    found=1;
	    break;
	}
	tmp=iter;
	++iter;
    }

    if(found) {
	blockList.remove(*block);
	delete block;
	block=NULL;
	return tmp;
    } else {
	return NULL;
    }
}

CBlock *CGif::moveBlockUp(CBlock *block)
{
    int found=0;

  // Check for the header
    if(block->getType() == TYPE_GLOBAL) {
	errMsg("CGif::moveBlockUp: Cannot move the Global block");
	return NULL;
    }

  // Find the block by parsing the list (inefficient)
    CBlock *tmp=blockList.first();;
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if((CBlock *)iter == block) {
	    found=1;
	    break;
	}
	tmp=iter;
	++iter;
    }
    
    if(found) {
	if(tmp->getType() == TYPE_GLOBAL) {
	    errMsg("CGif::moveBlockUp: Already at the top");
	    return NULL;
	}
	blockList.remove(*block);
	blockList.insertBefore(*block,*tmp);
	return block;
    } else {
	return NULL;
    }
}

CBlock *CGif::moveBlockDown(CBlock *block)
{
    int found=0;

  // Check for the header
    if(block->getType() == TYPE_GLOBAL) {
	errMsg("CGif::moveBlockDown: Cannot move the Global block");
	return NULL;
    }
  // Check if at the end
    if(block == blockList.last()) {
	errMsg("CGif::moveBlockDown: Already at the bottom");
	return NULL;
    }

  // Find the block by parsing the list (inefficient)
    CBlock *tmp=blockList.last();;
    tsDLIterBD<CBlock> iter(blockList.last());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if((CBlock *)iter == block) {
	    found=1;
	    break;
	}
	tmp=iter;
	--iter;
    }

    if(found) {
	blockList.remove(*block);
	blockList.insertAfter(*block,*tmp);
	return block;
    } else {
	return NULL;
    }
}

CBlock *CGif::addComment(CBlock *block)
{
    CComment *comment=new CComment(this);
    if(!comment) {
	errMsg("CGif::addComment: Cannot allocate Comment");
	return NULL;
    }
    blockList.insertAfter(*(CBlock *)comment,*block);

    return (CBlock *)comment;
}

// Returns the Control and NULLs the pointer.  Is called by Image to
// transfer the Control to the Image.
CControl *CGif::moveControl(void)
{
    CControl *temp=control;
    control=NULL;
    return temp;
}

// Gets the byte order for this client (No longer used to set the
// ClientByteOrder, which is always MSBFirst)
int CGif::getClientByteOrder()
{
    short i=1;
    
    return (*(char*)&i == 1) ? LSBFirst : MSBFirst;
}

// Sets the options currently in the main program
void CGif::setGraphicsOptions(void)
{
    setBgMode(::bgMode);
    dataMode=::dataMode;
}

void CGif::setBgMode(const int bgModeIn)
{
    bgMode=bgModeIn;

  // Find the images and redo the transparent color
    CBlock *tmp=blockList.first();;
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    if(image->getTransparentColorFlag()) {
		image->writeImageData();
	    }
	}
	tmp=iter;
	++iter;
    }
}

void CGif::setTransparentColorFlag(Boolean flag)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setTransparentColorFlag(flag);
	}
	++iter;
    }
}

void CGif::setInterlace(Boolean flag)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setInterlace(flag);
	}
	++iter;
    }
}

void CGif::setLocalColorTableFlag(Boolean flag)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setLocalColorTableFlag(flag);
	}
	++iter;
    }
}

void CGif::setTransparentIndex(char * value)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setTransparentIndex(value);
	}
	++iter;
    }
}

void CGif::setDelayTime(char * value)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setDelayTime(value);
	}
	++iter;
    }
}

void CGif::setLeftOffset(char * value)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setLeftOffset(value);
	}
	++iter;
    }
}

void CGif::setTopOffset(char * value)
{
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->setTopOffset(value);
	}
	++iter;
    }
}

// Sets the logical screen width.  Checks to be sure all images fit.
// Returns SGA_ERROR if it can't use the given values.
int CGif::setLogicalScreenWidth(int width)
{
    int minWidth=0;
    int error=0;

  // Find the minimums
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    int iWidth=image->getWidthExtent();
	    if(iWidth > minWidth) minWidth=iWidth;
	}
	++iter;
    }

  // Change to the greater of the specified value or the minimum
    if(width >= minWidth) {
	LogicalScreenWidth=width;
    } else {
	LogicalScreenWidth=minWidth;
	error=1;
    }

    return error?SGA_ERROR:SGA_OK;
}

// Sets the logical screen height.  Checks to be sure all images fit.
// Returns SGA_ERROR if it can't use the given values.
int CGif::setLogicalScreenHeight(int height)
{
    int minHeight=0;
    int error=0;

  // Find the minimums
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    int iHeight=image->getHeightExtent();
	    if(iHeight > minHeight) minHeight=iHeight;
	}
	++iter;
    }

  // Change to the greater of the specified value or the minimum
    if(height >= minHeight) {
	LogicalScreenHeight=height;
    } else {
	LogicalScreenHeight=minHeight;
	error=1;
    }

    return error?SGA_ERROR:SGA_OK;
}

// Merge the blocks from another GIF.
int CGif::mergeGif(CGif &gifIn, CBlock *block)
{
  // Get all but the first (global) block and insert them in this.blockList
#if DEBUG_MERGE
    print("CGif::mergeGif: countIn=%d count=%d\n",
      gifIn.getBlockCount(),blockList.count());
#endif
    while(gifIn.getBlockCount() > 1) {
	CBlock *curBlock=gifIn.getBlock(2,True);
#if DEBUG_MERGE > 1
	print("                countIn=%d count=%d\n",
	  gifIn.getBlockCount(),blockList.count());
	print("This blockList:\n");
	dumpBlockList();
	print("gifIn blockList:\n");
	gifIn.dumpBlockList();
#endif
	if(curBlock) {
	    curBlock->setGif(this);
	    blockList.insertAfter(*curBlock,*block);
	    block=curBlock;
	} else {
	    errMsg("CGif::MergeGif: Encountered NULL block");
	    exit(1);
	}
    }

  // Reset images
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    int number=1;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *curBlock=iter;
	    CImage *image=(CImage *)curBlock;
	    image->setImageNumber(number++);
	    image->writeImageData();
	    nImages++;
	}
	++iter;
    }

  // Fix the size if necessary
    if(gifIn.getLogicalScreenWidth() > LogicalScreenWidth)
      LogicalScreenWidth=gifIn.getLogicalScreenWidth();
    if(gifIn.getLogicalScreenHeight() > LogicalScreenHeight)
      LogicalScreenHeight=gifIn.getLogicalScreenHeight();
    
    return SGA_OK;
}

// Sets a new value of nColors
void CGif::setNColors(int nColorsIn)
{
    BitsPerPixel=0;
    while((nColorsIn&1) != nColorsIn) nColorsIn>>=1,BitsPerPixel++;
    SizeOfGlobalColorTable=BitsPerPixel-1;
    nColors=ColorTableEntries=1<<BitsPerPixel;
    BitMask=ColorTableEntries-1;
}

// Set the colors for an index in the color table.  The color
// arguments are those from an XColor structure.
int CGif::setColor(int i, int red, int green, int blue, Boolean free)
{
    Pixel blackPixel=BlackPixel(display,screen);
    Colormap colormap=DefaultColormap(display,screen);
    XColor xColor;
    int error=0;
    

  // Check index is in range
    if(i < 0 || i >= nColors) {
	return SGA_ERROR;
    }

  // Set the internal variables
    Red[i]=red>>8;
    Green[i]=green>>8;
    Blue[i]=blue>>8;
    xColor.red=Red[i]<<8;
    xColor.green=Green[i]<<8;
    xColor.blue=Blue[i]<<8;
    xColor.flags=DoRed|DoGreen|DoBlue;

  // Free the old color
    if(free && pixels[i] != blackPixel)
      XFreeColors(display,colormap,&pixels[i],1,0);

  // Allocate the  new color
    if(!XAllocColor(display,colormap,&xColor)) { 
#if 0
	errMsg("CGif::setColor: Cannot allocate new color for "
	  "index %d [%2x,%2x,%2x]",
	  i,Red[i],Green[i],Blue[i]);
#endif
	xColor.pixel=blackPixel;
	error=1;
    }
    pixels[i]=xColor.pixel;

  // Redo the images
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->writeImageData();
	}
	++iter;
    }

    return error?SGA_ERROR:SGA_OK;
}

// Check if the color is used
Boolean CGif::isUsedColor(Byte index)
{

    if(index >= nColors) return False;

  // Look at each image
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    if(image->isUsedColor(index)) {
		return True;
	    }
	}
	++iter;
    }

    return False;
}

// Change the color indices in the image blocks according to a
// translation table.
void CGif::translateColors(Byte *table)
{
  // Look at each image
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *block=iter;
	    CImage *image=(CImage *)block;
	    image->translateColors(table);
	}
	++iter;
    }
}

// Change the indices to the color which is the closest match in the
// color table of the reference.
int CGif::matchColors(CGif &reference)
{
    int r2,g2,b2,sumsq;

#if DEBUG_MERGE
    print("CGif::matchColors:\n");
#endif

  // Find the match by minimizing the sum of the squares of the r,g
  // and b values
    Byte *table=new Byte[256];
    for(int i=0; i < nColors; i++) {
	int fit=INT_MAX;
	table[i]=0;
	for(int j=0; j < reference.getNColors(); j++) {
	    r2=reference.getRed(j)-Red[i];
	    g2=reference.getGreen(j)-Green[i];
	    b2=reference.getBlue(j)-Blue[i];
	    sumsq=r2*r2+g2*g2+b2*b2;
	    if(sumsq < fit) {
		fit=sumsq;
		table[i]=j;
	    }
	  // Quit if we have an exact match
	    if(!sumsq) break;
	}
#if DEBUG_MERGE
	int k=table[i];
	print("  %3d %2x %2x %2x  %3d %2x %2x %2x\n",
	  i,Red[i],Green[i],Blue[i],
	  k,
	  reference.getRed(k),reference.getGreen(k),reference.getBlue(k));
#endif
    }

  // Set the new colors
    translateColors(table);

  // Merge the blocks
    int status=reference.mergeGif(*this,selectedBlock);
    if(status != SGA_OK) {
	errMsg("CGif::matchColors: Cannot merge %s\n",
	  fileName?fileName:"New GIF");
    }

  // Always suceeds
    delete [] table;
    return SGA_OK;
}

// Replace colors as needed in the reference color table and change
// the indices appropriately.  Do not add.
int CGif::replaceColors(CGif &reference)
{
    int refNColors=reference.getNColors();
    Boolean used[256],refUsed[256],changed[256];
    Byte newRed[256],newGreen[256],newBlue[256];
    int nBadColors=0,nHandledColors=0;
    int error=0;
    int i,j,found;
    int status;

#if DEBUG_MERGE
    print("CGif::replaceColors:\n");
#endif

  // First find the unused colors and initialize
    int nUsedColors=0;
    for(i=0; i < nColors; i++) {
	used[i]=isUsedColor(i);
	if(used[i]) nUsedColors++;
    }
    for(i=0; i < refNColors; i++) {
	changed[i]=False;
	newRed[i]=reference.getRed(i);
	newGreen[i]=reference.getGreen(i);
	newBlue[i]=reference.getBlue(i);
	refUsed[i]=reference.isUsedColor(i);
    }

  // Look for the colors in the reference
    Byte *table=new Byte[256];
    for(i=0; i < nColors; i++) {
	table[i]=0;
      // Don't do anything if the color is not used anyway
	if(!used[i]) continue;
      // First check for an exact match of a used color
	found=0;
	for(j=0; j < refNColors; j++) {
	    if(!refUsed[j]) continue;
	    if(newRed[j] == Red[i] &&
	      newGreen[j] == Green[i] &&
	      newBlue[j] == Blue[i]) {
	      // Found
		table[i]=j;
		refUsed[j]=True;
		nHandledColors++;
		found=1;
		break;
	    }
	}
	if(found) continue;
      // Next look for unused colors
	found=0;
	for(j=0; j < refNColors; j++) {
	    if(!refUsed[j]) {
	      // Found
		table[i]=j;
		refUsed[j]=True;
		newRed[j]=Red[i];
		newGreen[j]=Green[i];
		newBlue[j]=Blue[i];
		changed[j]=True;
		nHandledColors++;
		found=1;
		break;
	    }
	}
	if(!found) {
	    errMsg("CGif::replaceColors:\n"
	      "  The color table size is %d\n"
	      "  There are %d new colors that are used\n"
	      "  Only %d of %d colors can be matched or replaced\n"
	      "  %d colors cannot be handled",
	      refNColors,nUsedColors,nHandledColors,refNColors,
	      nUsedColors-nHandledColors);
	    error=1;
	    goto CLEANUP;
	}
    }

  // Change the colors in reference
    for(i=0; i < refNColors; i++) {
	if(changed[i]) {
	    status=reference.setColor(i,
	      (int)newRed[i]<<8,(int)newGreen[i]<<8,(int)newBlue[i]<<8);
	    if(status != SGA_OK) nBadColors++;
	}
    }
    if(nBadColors) {
	errMsg("CGif::replaceColors: Could not allocate %d colors",nBadColors);
    }
    reference.resetBackgroundColor();

  // Set the new colors in this
    translateColors(table);

  // Merge the blocks
    status=reference.mergeGif(*this,selectedBlock);
    if(status != SGA_OK) {
	errMsg("CGif::replaceColors: Cannot merge %s\n",
	  fileName?fileName:"New GIF");
	error=1;
	goto CLEANUP;
    }

CLEANUP:
#if DEBUG_MERGE
    for(i=0; i < nColors; i++) {
	int k=table[i];
	print("  %3d %2x %2x %2x %s  %3d %2x %2x %2x %s %s\n",
	  i,Red[i],Green[i],Blue[i],
	  used[i]?"Used  ":"Unused",
	  k,
	  newRed[k],newGreen[k],newBlue[k],
	  refUsed[k]?"Used  ":"Unused",
	  changed[k]?"Changed  ":"Unchanged");
    }
#endif
    delete [] table;
    return error?SGA_ERROR:SGA_OK;
}

// Add colors as needed in the reference color table and change the
// indices appropriately.  Do not replace.
int CGif::addColors(CGif &reference)
{
    int refNColors=reference.getNColors();
    int newNColors;
    Boolean used[256];
    int nBadColors=0;
    int i,next;

#if DEBUG_MERGE
    print("CGif::addColors:\n");
#endif

  // First find the unused colors
    int nUsedColors=0;
    for(i=0; i < nColors; i++) {
	used[i]=isUsedColor(i);
	if(used[i]) nUsedColors++;
    }

  // Return if there are too many new colors
    if(refNColors+nUsedColors > 256) {
	errMsg("CGif::addColors:\n"
	  "  The color table size is %d\n"
	  "  There are %d new colors that are used\n"
	  "  There would be more than the allowed 256 colors",
	  refNColors,nUsedColors);
	return SGA_ERROR;
    }

  // Calculate the new number of colors.  Must be power of 2.
    newNColors=refNColors;
    while(refNColors+nUsedColors > newNColors) newNColors*=2;

  // Add the colors
    next=refNColors;
    Byte *table=new Byte[256];
    for(i=0; i < nColors; i++) {
	table[i]=0;
      // Don't do anything if the color is not used anyway
	if(!used[i]) continue;
	table[i]=next;
	next++;
#if DEBUG_MERGE
	if(i == 0) print("newNColors=%d\n",newNColors);
	int k=table[i];
	print("  %3d %2x %2x %2x  %3d %2x %2x %2x ->  %2x %2x %2x\n",
	  i,Red[i],Green[i],Blue[i],
	  k,
	  reference.getRed(k),reference.getGreen(k),reference.getBlue(k),
	  Red[i],Green[i],Blue[i]);
#endif
    }

  // Change the colors in reference
    reference.setNColors(newNColors);
    next=refNColors;
    for(i=0; i < nColors; i++) {
	if(!used[i]) continue;
      // Set the colors without freeing them first
	int status=reference.setColor(next,
	  (int)Red[i]<<8,(int)Green[i]<<8,(int)Blue[i]<<8,
	  False);
	if(status != SGA_OK) nBadColors++;
	next++;
    }
    if(nBadColors) {
	errMsg("CGif::addColors: Could not allocate %d colors",nBadColors);
    }
    reference.resetBackgroundColor();

  // Set the new colors in this
    translateColors(table);

  // Merge the blocks
    int status=reference.mergeGif(*this,selectedBlock);
    if(status != SGA_OK) {
	errMsg("CGif::addColors: Cannot merge %s\n",
	  fileName?fileName:"New GIF");
    }

    delete [] table;
    return SGA_OK;
}

// First try to replace, then add colors as needed in the reference
// color table and change the indices appropriately.
int CGif::addReplaceColors(CGif &reference)
{
    int refNColors=reference.getNColors();
    Boolean used[256],refUsed[256],changed[256];
    Byte newRed[256],newGreen[256],newBlue[256];
    int nBadColors=0,nHandledColors=0,nNewColors,newNColors;
    int error=0;
    int i,j,found,next,curi;
    int status;

#if DEBUG_MERGE
    print("CGif::addReplaceColors:\n");
#endif

  // First find the unused colors and initialize
    int nUsedColors=nNewColors=0;
    for(i=0; i < nColors; i++) {
	used[i]=isUsedColor(i);
	if(used[i]) nUsedColors++;
    }
    for(i=0; i < 256; i++) {
	changed[i]=False;
	newRed[i]=reference.getRed(i);
	newGreen[i]=reference.getGreen(i);
	newBlue[i]=reference.getBlue(i);
	if(i < refNColors) 
	  refUsed[i]=reference.isUsedColor(i);
	else
	  refUsed[i]=False;
    }

  // Look for the colors in the reference
    Byte *table=new Byte[256];
    curi=0;
    for(i=0; i < nColors; i++) {
	curi=i;
	table[i]=0;
      // Don't do anything if the color is not used anyway
	if(!used[i]) continue;
      // First check for an exact match of a used color
	found=0;
	for(j=0; j < refNColors; j++) {
	    if(!refUsed[j]) continue;
	    if(newRed[j] == Red[i] &&
	      newGreen[j] == Green[i] &&
	      newBlue[j] == Blue[i]) {
	      // Found
		table[i]=j;
		refUsed[j]=True;
		nHandledColors++;
		found=1;
		break;
	    }
	}
	if(found) continue;
      // Next look for unused colors
	found=0;
	for(j=0; j < refNColors; j++) {
	    if(!refUsed[j]) {
	      // Found
		table[i]=j;
		refUsed[j]=True;
		newRed[j]=Red[i];
		newGreen[j]=Green[i];
		newBlue[j]=Blue[i];
		changed[j]=True;
		nHandledColors++;
		found=1;
		break;
	    }
	}
	if(found) continue;
      // Next try to add colors
	nNewColors=nUsedColors-nHandledColors;
	if(refNColors+nNewColors > 256) {
	    errMsg("CGif::addReplaceColors:\n"
	      "  The color table size is %d\n"
	      "  There are %d new colors that are used\n"
	      "  %d colors can be matched or replaced\n"
	      "  There there is not enough room for the remaining %d colors",
	      refNColors,nUsedColors,nHandledColors,nNewColors);
	    error=1;
	    goto CLEANUP;
	}
	break;
    }

  // Add new colors
    if(nNewColors) {
      // Calculate the new number of colors.  Must be power of 2.
	newNColors=refNColors;
	while(refNColors+nNewColors > newNColors) newNColors*=2;
	
      // Add the colors
	next=refNColors;
	for(i=curi; i < nColors; i++) {
	    table[i]=0;
	  // Don't do anything if the color is not used anyway
	    if(!used[i]) continue;
	    table[i]=next;
	    newRed[next]=Red[i];
	    newGreen[next]=Green[i];
	    newBlue[next]=Blue[i];
	    changed[next]=True;
	    refUsed[next]=True;
	    next++;
	}
    } else {
	newNColors=refNColors;
    }

  // Change the colors in reference
    if(nNewColors) {
	reference.setNColors(newNColors);
    }
    for(i=0; i < newNColors; i++) {
	if(changed[i]) {
	    status=reference.setColor(i,
	      (int)newRed[i]<<8,(int)newGreen[i]<<8,(int)newBlue[i]<<8);
	    if(status != SGA_OK) nBadColors++;
	}
    }
    if(nBadColors) {
	errMsg("CGif::replaceColors: Could not allocate %d colors",nBadColors);
    }
    reference.resetBackgroundColor();

#if 0
  // Change the colors in reference
    if(nNewColors) reference.setNColors(newNColors);
    for(i=0; i < nColors; i++) {
	if(changed[i]) {
	    status=reference.setColor(table[i],
	      (int)newRed[i]<<8,(int)newGreen[i]<<8,(int)newBlue[i]<<8);
	    if(status != SGA_OK) nBadColors++;
	}
    }
    if(nBadColors) {
	errMsg("CGif::addReplaceColors: Could not allocate %d colors",nBadColors);
    }
    reference.resetBackgroundColor();
#endif

  // Set the new colors in this
    translateColors(table);

  // Merge the blocks
    status=reference.mergeGif(*this,selectedBlock);
    if(status != SGA_OK) {
	errMsg("CGif::addReplaceColors: Cannot merge %s\n",
	  fileName?fileName:"New GIF");
	error=1;
	goto CLEANUP;
    }

CLEANUP:
#if DEBUG_MERGE
    print("  nColors=%d nUsedColors=%d refNColors=%d nHandledColors=%d"
      " newNColors=%d\n",
      nColors,nUsedColors,refNColors,nHandledColors,newNColors);
    for(i=0; i < nColors; i++) {
	int k=table[i];
	print("  %3d %2x %2x %2x %s  %3d %2x %2x %2x %s %s\n",
	  i,Red[i],Green[i],Blue[i],
	  used[i]?"Used  ":"Unused",
	  k,
	  newRed[k],newGreen[k],newBlue[k],
	  refUsed[k]?"Used  ":"Unused",
	  changed[k]?"Changed  ":"Unchanged");
    }
#endif
    delete [] table;
    return error?SGA_ERROR:SGA_OK;
}


// Reduce the color table size to the minimum needed to hold the used
// colors and translate the used colors
int CGif::trimColorTable(void)
{
    Boolean used[256],changed[256];
    Byte newRed[256],newGreen[256],newBlue[256];
    int nUsed,newNColors,nBadColors;
    int i;

#if DEBUG_MERGE
    print("CGif::trimColorTable:\n");
#endif

  // First find the unused colors
    nUsed=0;
    for(i=0; i < nColors; i++) {
	used[i]=isUsedColor(i);
	if(used[i]) nUsed++;
    }
    newNColors=2;
    while(nUsed > newNColors) newNColors*=2;
    if(newNColors >= nColors) {
	errMsg("CGif::trimColorTable:\n"
	  "  The color table is already as small as it can be");
	return SGA_ERROR;
    }
    
  // Look for the unused colors
    Byte *table=new Byte[256];
    int next=0;
    for(i=0; i < nColors; i++) {
	table[i]=0;
	if(!used[i]) continue;
	table[i]=next;
	changed[next]=True;
	newRed[next]=Red[i];
	newGreen[next]=Green[i];
	newBlue[next]=Blue[i];
	next++;
#if DEBUG_MERGE
	int k=table[i];
	print("  %3d %2x %2x %2x  %3d %2x %2x %2x\n",
	  i,Red[i],Green[i],Blue[i],
	  k,newRed[k],newGreen[k],newBlue[k]);
#endif
    }

  // Change the colors
    nBadColors=0;
    for(i=0; i < nUsed; i++) {
	if(changed[i]) {
	    int status=setColor(i,
	      (int)newRed[i]<<8,(int)newGreen[i]<<8,(int)newBlue[i]<<8);
	    if(status != SGA_OK) nBadColors++;
	}
    }
    if(nBadColors) {
	errMsg("CGif::trimColorTable: Could not allocate %d colors",nBadColors);
    }
    setNColors(newNColors);
    resetBackgroundColor();

  // Set the new colors in this
    translateColors(table);

  // Reset images
    tsDLIterBD<CBlock> iter(blockList.first());
    tsDLIterBD<CBlock> eol;
    int number=1;
    while(iter != eol) {
	if(iter->getType() == TYPE_IMAGE) {
	    CBlock *curBlock=iter;
	    CImage *image=(CImage *)curBlock;
	    image->setImageNumber(number++);
	    image->writeImageData();
	}
	++iter;
    }

    delete [] table;
    return SGA_OK;
}
