/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// Header file for SGA classes

#define ST_START 1
#define ST_FILEREAD 2
#define ST_HEADERDONE 3
#define ST_COLORDONE 4
#define ST_ALLDONE 5

// Fix static stringTable in CBlock.cc if these are changed
#define TYPE_GLOBAL 0
#define TYPE_IMAGE 1
#define TYPE_CONTROL 2
#define TYPE_PLAINTEXT 3
#define TYPE_APPLICATION 4
#define TYPE_COMMENT 5
#define TYPE_UNKNOWN 6
#define N_STRING_TYPES TYPE_UNKNOWN+1
#define LABEL_SIZE 13

#define SGA_OK 0
#define SGA_ERROR 1

#define NEXTBYTE (*ptr++)

#define IMAGESEP 0x2c
#define TRAILER 0x3b
#define EXTENSION 0x21
#define GRAPHCONTROL 0xf9
#define PLAINTEXT 0x01
#define APPLICATION 0xff
#define COMMENT 0xfe
#define BLOCKTERM 0x00
#define MASK0   0x01
#define MASK210 0x07
#define MASK432 0x1c
#define MASK654 0x70
#define MASK3   0x08
#define MASK6   0x40
#define MASK7   0x80

#define TEXT_MODE    0
#define GRAPHIC_MODE 1

#define TEXT_AREA 0
#define DRAW_AREA 1

#define BACKGROUND_GIF   0
#define BACKGROUND_WHITE 1
#define BACKGROUND_BLACK 2

#define DA_MARGIN 10

#include "sga.h"

class CBlock;
class CControl;
class CGlobal;

typedef unsigned char Byte;

// Class to hold the GIF.

/**************************************************************************/
/**************************** CGif ****************************************/
/**************************************************************************/

class CGif
{
    friend class CGlobal;
    
  public:
    CGif(void);
    CGif(int nColorsIn);
    ~CGif(void);
    const char *getShortName(void) const;
    const char *getLongName(void) const { return fileName; }
    void setLongName(const char *fileNameIn);
    int load(char *fileNameIn);
    int writeFile(const char *fileNameIn);
    int checkRemainingSize(const int bytes) const
      { return (ptr+bytes <= eof)?True:False; }
    Byte *getPtr(void) const { return ptr; };
    int getScreenDepth(void) const { return ScreenDepth; }
    int getNColors(void) const { return nColors; }
    void setNColors(int nColorsIn);
    const Pixel *getPixels(void) const { return pixels; }
    Byte getRed(int i) const { return Red[i]; }
    Byte getGreen(int i) const { return Green[i]; }
    Byte getBlue(int i) const { return Blue[i]; }
    int setColor(int i, int red, int green, int blue, Boolean free=True);
    Pixel getBackgroundColor(void) const { return BackgroundColor; }
    void resetBackgroundColor(void)
      { BackgroundColor=pixels[BackgroundColorIndex]; }
    int getBackgroundColorIndex(void) const { return BackgroundColorIndex; }
    int getBitsPerPixel(void) const { return BitsPerPixel; }
    int getLogicalScreenWidth(void) const { return LogicalScreenWidth; }
    int getLogicalScreenHeight(void) const { return LogicalScreenHeight; }
    int setLogicalScreenWidth(int width);
    int setLogicalScreenHeight(int height);
    int getBytesPerPixel(void) const { return BytesPerPixel; }
    int getBitMask(void) const { return BitMask; }
    Visual *getVisual(void) const { return visual; }
    CControl *moveControl(void);
    void setPtr(Byte *ptrIn) { ptr=ptrIn; };
    int updateListArea(void);
    int getClientByteOrder();
    int getDataMode(void) const { return dataMode; }
    void setDataMode(const int dataModeIn) { dataMode=dataModeIn; }
    int getBgMode(void) const { return bgMode; }
    void setBgMode(const int bgModeIn);
    void setGraphicsOptions(void);
    void printSummary(void) const;
    int getNextImageNumber(void) { return ++nImages; }
    int getNImages(void) { return nImages; }

    unsigned getBlockCount(void) const { return blockList.count(); }
    CBlock *getBlock(const int pos, Boolean remove=False);
    int getPos(const CBlock *block) const;
    CBlock *getFirstImage(void) const;
    CBlock *getLastImage(void) const;
    CBlock *getPrevImage(const CBlock *block) const;
    CBlock *getNextImage(const CBlock *block) const;

    void markEdited(void) { edited=True; }
    CBlock *deleteBlock(CBlock *block);
    CBlock *moveBlockUp(CBlock *block);
    CBlock *moveBlockDown(CBlock *block);
    CBlock *addComment(CBlock *block);

  // Routines for combining color tables
    Boolean isUsedColor(Byte index);
    void translateColors(Byte *table);
    int matchColors(CGif &reference);
    int replaceColors(CGif &reference);
    int addColors(CGif &reference);
    int addReplaceColors(CGif &reference);
    int mergeGif(CGif &gifIn, CBlock *block);
    int trimColorTable(void);
    
  // These routines set values for all images
    void setTransparentColorFlag(Boolean flag);
    void setInterlace(Boolean flag);
    void setLocalColorTableFlag(Boolean flag);
    void setTransparentIndex(char *value);
    void setDelayTime(char *value);
    void setLeftOffset(char *value);
    void setTopOffset(char *value);

  // Debug
    void dumpBlockList(void) const;
    void print1(const char *name) const;
    
  private:
    void cleanup(void);
    int openFile(char *fname);    
    int parseHeader(void);
    int parseGlobalColorTable(void);
    int allocateColors(void);
    int parseLocalColorTable(void);
    int parseRemainderOfFile(void);
    int parseExtension(void);

  private:
    CControl *control;
    int state;
    int dataMode;
    int bgMode;
    Boolean edited;
    int nImages;

    char fileName[PATH_MAX];
    int fileSize;
    Byte *raw;    // The bytes in the file
    Byte *ptr;    // Current position in raw data
    Byte *eof;    // Byte after the last character in the raw data

    int nColors;
    Byte Red[256], Green[256], Blue[256];
    Pixel pixels[256];

    tsDLList<CBlock> blockList;

  // From the header
    char GifType[7];                  // GIF Version
    int LogicalScreenWidth;           // Logical screen width
    int LogicalScreenHeight;          // Logical screen height
    Boolean GlobalColorTableFlag;     // Flag indicating global color table follows
    Boolean GlobalColorTableSortFlag; // Flag indicating global color table is sorted
    int ColorResolution;              // Color resolution
    int SizeOfGlobalColorTable;       // Size of Global Color Table from header
    int BitsPerPixel;                 // Bits per pixel (SizeOfGlobalColorTable - 1)
    int ColorTableEntries;            // Number of colors in global color table
    int BitMask;                      // AND mask for data size
    int BackgroundColorIndex;         // Background color index
    Pixel BackgroundColor;            // Background color pixel
    int PixelAspectRatio;             // Pixel aspect ratio
    
  // X Parameters (Order is significant for initialization)
    int ScreenDepth;                  // X Screen depth
    int XBitsPerPixel;                // X bits per pixel
    int XBitmapBitOrder;              // XImage bitmap bit order
    int XBitmapPad;                   // XImage bitmap pad
    int XBitmapUnit;                  // XImage bitmap unit
    int XByteOrder;                   // XImage byte order
    int BytesPerPixel;                // Bytes per pixel
    Visual *visual;                   // Visual

  // Machine parameters
    int ClientByteOrder;
};

/**************************************************************************/
/**************************** CBlock **************************************/
/**************************************************************************/

// Class for each block of the GIF.  These are nodes in the CGif
// blockList.

class CBlock : public tsDLNode<CBlock>
{
  public:
    CBlock(CGif *parent, int typeIn);
    virtual ~CBlock(void);
    char *stringType(void) const;
    int getType(void) const { return type; }
    CGif *getGif(void) const { return pGif; }
    void setGif(CGif *gif) { pGif=gif; }

  // Virtual functions
    virtual void storeSize(void)=0;
    virtual int write(FILE *fp)=0;
    virtual int parse(void)=0;
    virtual int displayData(void)=0;
    virtual int displayOptions(void)=0;
    
  protected:
    int storeData(void);
    int storeText(void);
    void writeWord(int word,FILE *fp) const;

  protected:
    int rawSize;    // Size of entire block
    int introSize;  // Size of bytes before the data blocks
    int textSize;
    CGif *pGif;
    int type;
    Byte *raw;
    char *text;
};

/**************************************************************************/
/**************************** CControl ************************************/
/**************************************************************************/

class CControl : public CBlock
{
    friend class CImage;
    
  public:
    CControl(CGif *parent);
    ~CControl(void);
    Byte getTransparentColorFlag(void) const { return TransparentColorFlag; }
    Byte getTransparentIndex(void) const { return TransparentIndex; }
    int getDelayTime(void) const { return DelayTime; }
    int getDisposalMethod(void) const { return DisposalMethod; }

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);

  private:
    Byte PackedByte;           // Packed byte
    Byte TransparentColorFlag; // Use Transparent color for the frame
    Byte TransparentIndex;     // Transparent color index for the frame
    int DelayTime;             // Delay time in 1/100 sec for the frame
    int DisposalMethod;        // Image disposal method
};

/**************************************************************************/
/**************************** CImage **************************************/
/**************************************************************************/

class CImage : public CBlock
{
    friend class CControl;

  public:
    CImage(CGif *parent);
    ~CImage(void);

    int getDelayTime(void) const { return control->DelayTime; }
    int getTransparentColorFlag(void) const
      { return control->TransparentColorFlag; }
    int getTransparentIndex(void) const { return control->TransparentIndex; }
    int getDisposalMethod(void) const { return control->DisposalMethod; }
    Boolean getInterlace(void) const { return Interlace; }
    int getLeftOffset(void) const { return LeftOffset; }
    int getTopOffset(void) const { return TopOffset; }
    int getWidthExtent(void) const { return LeftOffset+Width; }
    int getHeightExtent(void) const { return TopOffset+Height; }
    Pixmap getPixmap(void) const { return pixmap; }
    int getImageNumber(void) const { return imageNumber; }
    void setImageNumber(int number) { imageNumber=number; }
    int makePixmap(void);
    Boolean isUsedColor(Byte index);
    void translateColors(Byte *table);

    void updateOptionsArea(void) const;
    void setTransparentColorFlag(Boolean flag);
    void setInterlace(Boolean flag);
    void setLocalColorTableFlag(Boolean flag);
    void setTransparentIndex(char *value);
    void setDelayTime(char *value);
    void setLeftOffset(char *value);
    void setTopOffset(char *value);
    void setDisposalMethod(int method);

    void exposeCB(void);
    void resizeCB(void);

    void writeImageData(void);    // Elsewhere this is private

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);


    static char *getDisposalMethodString(int n);

  private:
    int readCode(void);
    void addToColorData(Byte Index);
    void addToImageData(Byte Index);

  private:
    CControl *control;                // Associated Control block
    Byte *raster;                     // Temporary storage
    Byte *imageData;                  // Data for the X image
    Byte *colorData;                  // Data for the color array
    int imageDataSize;                // The size of the XImage data
    int colorDataSize;                // The size of the color data
    XImage *image;                    // The X image
    GC gc;                            // Graphics context for drawing area
    Pixmap pixmap;                    // Pixmap for drawing area repaints
    Dimension pWidth,pHeight;         // Pixmap dimensions
    int imageNumber;                  // Image number in the CGif

  // The hash table used by the decompressor
    int *prefix;
    int *suffix;

  // An output array used by the decompressor
    int *outCode;

  // Parameters from the header (determined in the CGif)
    int nColors;
    const Pixel *pixels;
    int ScreenDepth;                  // X Screen depth
    int BytesPerPixel;                // Bytes per pixel
    int BitMask;                      // AND mask for data size

  // Local to an Image
    Boolean Interlace;                // Interlace flag
    int BitOffset;                    // Bit Offset of next code
    int XC, YC;                       // Output X and Y coords of current pixel
    int Pass;                         // Used by output routine if interlaced pic
    int OutCount;                     // Decompressor output 'stack count'

    int Width, Height;                // Image dimensions
    int LeftOffset, TopOffset;        // Image offsets from logical screen
    int BytesPerScanline;             // Bytes per scanline in output raster
    int CodeSize;                     // Code size, read from GIF header
    int InitCodeSize;                 // Starting code size, used during Clear
    int Code;                         // Value returned by readCode
    int MaxCode;                      // Limiting value for current code size
    int ClearCode;                    // GIF clear code
    int EOFCode;                      // GIF end-of-information code
    int CurCode, OldCode, InCode;     // Decompressor variables
    int FirstFree;                    // First free code, generated per GIF spec
    int FreeCode;                     // Decompressor, next free slot in hash table
    int FinChar;                      // Decompressor variable
    int ReadMask;                     // Code AND mask for current code size

  // Local color table
    int LocalColorTableFlag;          // Flag denoting a local color table
    int SizeOfLocalColorTable;        // Size of a local color table
    int LocalColorTableEntries;       // Number of colors in local color table
};

/**************************************************************************/
/**************************** CPlaintext **********************************/
/**************************************************************************/

class CPlaintext : public CBlock
{
  public:
    CPlaintext(CGif *parent);
    ~CPlaintext(void);

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);
};

/**************************************************************************/
/**************************** CApplication ********************************/
/**************************************************************************/

class CApplication : public CBlock
{
  public:
    CApplication(CGif *parent);
    ~CApplication(void);

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);

  private:
    char ApplicationIdentifier[9];
    char ApplicationAuthenticationCode[4];
};

/**************************************************************************/
/**************************** CComment ************************************/
/**************************************************************************/

class CComment : public CBlock
{
  public:
    CComment(CGif *parent);
    ~CComment(void);
    int storeText(void);     // Override

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);

    int setComment(char *value);

  private:
    char *comment;
    int commentSize;
};

/**************************************************************************/
/**************************** CGlobal *************************************/
/**************************************************************************/

class CGlobal : public CBlock
{
  public:
    CGlobal(CGif *parent);
    ~CGlobal(void);
    int storeText(void);     // Override

    virtual void storeSize(void);
    virtual int parse(void);
    virtual int write(FILE *fp);
    virtual int displayData(void);
    virtual int displayOptions(void);

    void updateOptionsArea(void) const;
    void setBackgroundColorIndex(char *value);
    void setLogicalScreenWidth(char *value);
    void setLogicalScreenHeight(char *value);
};
