/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// Implementation of CXpm


#define DEBUG_XPM 0
#define DEBUG_COLOR 0     // Use a small XPM

#define SAFE(x) (((x) < nColors)?(x):0)

#include <math.h>
#include "CXpm.h"

// Function prototypes

static void print(const char *fmt, ...);

// KE: The following is the prototype for _XGetBitsPerPixel, which is
// a convenience function found in the MIT distribution in
// /opt/local/share/src/R5/mit/lib/X/XImUtil.c (at the APS)
extern "C" {
int _XGetBitsPerPixel(Display *dpy, int depth);
}

// Utility for printing that works on both UNIX and WIN32
/**************************** print ***************************************/
static void print(const char *fmt, ...)
{
    va_list vargs;
    static char lstring[1024];  /* DANGER: Fixed buffer size */
    
    va_start(vargs,fmt);
    vsprintf(lstring,fmt,vargs);
    va_end(vargs);
    
    if(lstring[0] != '\0') {
#ifdef WIN32
	lprintf("%s",lstring);
#else
	printf("%s",lstring);
#endif
    }
}

/**************************************************************************/
/**************************** CXpm ****************************************/
/**************************************************************************/

CXpm::CXpm(Display *displayIn, char **xpmIn, Pixel bg, XpmErrMsg errMsgIn) :
    errMsg(errMsgIn),
    xpm(xpmIn),
    transparentColorFlag(False),
    transparentColorIndex(0),
    pixmap((Pixmap)NULL),
    imageData(NULL),
    colorData(NULL),
    image(NULL),
    gc(NULL),
    backgroundColor(bg),
    display(displayIn),
    screen(DefaultScreen(display)),
    screenDepth(DefaultDepth(display,screen)),
    XBitsPerPixel(_XGetBitsPerPixel(display,screenDepth)),
    XBitmapBitOrder(BitmapBitOrder(display)),
    XBitmapPad(BitmapPad(display)),
    XBitmapUnit(BitmapUnit(display)),
    XByteOrder(ImageByteOrder(display)),
    visual(DefaultVisual(display,screen)),
    bytesPerPixel(XBitsPerPixel/8),
    clientByteOrder(MSBFirst)
{
    int status=XPM_OK;

  // Make a graphics context
    XGCValues gcvalues;
    gcvalues.background=WhitePixel(display,screen);
    gcvalues.foreground=BlackPixel(display,screen);
    gc=XCreateGC(display,RootWindow(display,screen),
      GCBackground|GCForeground,&gcvalues);

  // Parse the XPM
    status=parse();
    if(status != XPM_OK) {
	goto RETURN;
    }

  // Write the array data to the image data
    writeImageData();

  // Make the pixmap
    status=redraw();
    if(status != XPM_OK) {
	goto RETURN;
    }

  RETURN:
#if DEBUG_XPM
    print("CXpm::CXpm: status=%s\n",
      status==XPM_OK?"XPM_OK":"XPM_ERROR");
    printSummary();
#endif
    return;
}

CXpm::~CXpm(void)
{
  // Delete the color array data
    if(colorData) {
	delete [] colorData;
	colorData=NULL;
    }
  // Delete the image
    if(image) {
      // Free the memory we allocated to insure it is done with the
      // same routines that allocated it.  Can be a problem on WIN32
      // if X frees it.
	if(image->data) free(image->data);
	image->data=NULL;
      // Then destroy the image
	XDestroyImage(image);
	image=NULL;
    }
  // Free the gc
    if(gc) {
	XFreeGC(display,gc);
	gc=NULL;
    }
  // Free the pixmap
    if(pixmap) {
	XFreePixmap(display,pixmap);
	pixmap=(Pixmap)NULL;
    }
}


int CXpm::parse(void)
{
    char *line,*tokens,*value;
    int i,j,len;
    int nTokens;
    UChar index;
    int status;
    int found;

    XC=0;
    YC=0;

  // XPM
  // static char *<variable-name>[]={
  //   <Values>
  //   <Colors>
  //   <Pixels>
  //   <Extensions>
  // };

#if DEBUG_XPM > 1
    print("CXpm::parse:  xpm[0]=%s\n",xpm[0]?xpm[0]:"NULL");
#endif

  // Values
    width=height=nColors=cpp=xhot=yhot=0;
    XPMEXT=NULL;
    line=xpm[0];
    len=strlen(line);
    tokens=new char[len+1];
    if(!tokens) {
	errMsg("CXpm::parse: Not enough memory for values");
	return XPM_ERROR;
    }
    strcpy(tokens,line);
    nTokens=0;
    for(i=0; i < 7; i++) {
	if(i) {
	    value=strtok(NULL," \t");
	} else {
	    value=strtok(tokens," \t");
	}
#if DEBUG_XPM > 1
	print("  value=%s\n",value?value:"NULL");
#endif
	if(!value) break;
	switch(i) {
	case 0:
	    width=atoi(value);
	    nTokens++;
	    break;
	case 1:
	    height=atoi(value);
	    nTokens++;
	    break;
	case 2:
	    nColors=atoi(value);
	    nTokens++;
	    break;
	case 3:
	    cpp=atoi(value);
	    nTokens++;
	    break;
	case 4:
	    xhot=atoi(value);
	    nTokens++;
	    break;
	case 5:
	    yhot=atoi(value);
	    nTokens++;
	    break;
	case 6:
	    XPMEXT=value;
	    nTokens++;
	    break;
	}
    }
    if(tokens) delete [] tokens;
    if(nTokens < 4) {
	errMsg("CXpm::parse: Need 4 values, got %d\n",nTokens);
	return XPM_ERROR;
    }
  // We only handle cpp=1 and use a translation table
    if(cpp != 1) {
	errMsg("CXpm::parse: Can only handle 1 characters per pixel [cpp],"
	  " got %d\n",cpp);
	return XPM_ERROR;
    }
  // We only 256 colors or fewer
    if(nColors > 256) {
	errMsg("CXpm::parse: Can only handle 256 or fewer colors,"
	  " got %d\n",nColors);
	return XPM_ERROR;
    }

  // We use a translation table instead of hashing.  Inefficient but easy.
  // Initialize the translation table
    for(i=0; i < 256; i++) {
	colorIndex[i]=0;
	colorChar[i]=0;
	*colorDef[i]='\0';
    }

  // Colors
    for(j=0; j < nColors; j++) {
	line=xpm[1+j];
#if DEBUG_XPM > 1
	print("CXpm::parse: xpm[%d]=%s\n",j+1,line?line:"NULL");
#endif
	index=line[0];
	colorIndex[index]=j;
	colorChar[j]=index;
	len=strlen(line);
	tokens=new char[len];     // Not len+1
	if(!tokens) {
	    errMsg("CXpm::parse: Not enough memory for colors");
	    return XPM_ERROR;
	}
	strcpy(tokens,line+1);
	value=strtok(tokens," \t");
	found=0;
      // We will look for c followed by a color
	while(value != NULL) {
	    if(!strcmp(value,"c")) {
	      // The next token should be the color
		found=1;
		value=strtok(NULL," \t");
		if(!value) {
		    errMsg("CXpm::parse: No color value found for keyword %c"
		      " for color %d",value,j);
		}
	      // Save the color definition string
		strncpy(colorDef[j],value,MAX_COLOR_STRING_LEN);
		colorDef[i][MAX_COLOR_STRING_LEN-1]='\0';
		break;
	    } else {
	      // Read and skip the color
		value=strtok(NULL," \t");
		if(!value) {
		    errMsg("CXpm::parse: No color value found for keyword %c"
		      " for color %d",value,j);
		    return XPM_ERROR;
		}
	      // Save the color definition string
	      // Will be used in case c is not found
		strncpy(colorDef[j],value,MAX_COLOR_STRING_LEN);
		colorDef[i][MAX_COLOR_STRING_LEN-1]='\0';
	    }
	    value=strtok(NULL," \t");
	}
	if(!found) {
	    errMsg("CXpm::parse: No color value found for color %d",j);
	    return XPM_ERROR;
	}
	if(tokens) delete [] tokens;
	parseColor(colorDef[j],j,Red+j,Green+j,Blue+j);
    }
    
  // Allocate the colors
    status=allocateColors();

  // Allocate the X Image
  // (Used to be switch statement on ScreenDepth here)
    colorDataSize=width*height;
    imageDataSize=bytesPerPixel*colorDataSize;

  // Allocate the data for the XImage
    if(!imageData) {
	imageData=(UChar *)malloc(imageDataSize);
	if(!imageData) {
	    errMsg("CXpm::parse: Not enough memory for XImage");
	    return XPM_ERROR;
	}
    }

  // Allocate the XImage
    if(!image) {
	image=XCreateImage(display,visual,
	  screenDepth,ZPixmap,0,
	  (char*)imageData,width,height,32,0);
	if(!image) {
	    errMsg("CXpm::parse: Unable to create XImage");
	    return XPM_ERROR;
	}
	bytesPerScanline=image->bytes_per_line;
    }

  // Fix the byte order for Exceed, possibly others
    image->byte_order=clientByteOrder;

  // Allocate the color array data
    if(!colorData) {
	colorData=new UChar[colorDataSize];
	if(!colorData) {
	    errMsg("CXpm::parse: Not enough memory for color array");
	    return XPM_ERROR;
	}
    }

  // Fill in the color data array
    XC=YC=0;
    for(j=0; j < height; j++) {
	line=xpm[1+nColors+j];
	for(i=0; i < width; i++) {
	  // The following is the index in the color table
	  // corresponding to the character at this position
	    index=colorIndex[line[i]&255];
#if DEBUG_COLOR
	    print(" line[i]=%c index=%d\n",line[i],index);
#endif
	    addToColorData(index);
	}
    }

#if DEBUG_XPM > 1
  // Dump all the arrays
    for(i=0; i < 256; i++) {
	print("%3d colorIndex=%d char=%c\n",
	  i,colorIndex[i],
	  (i > 31 && i < 127)?i:'.');
    }
    print("\n");
    for(j=0; j < height; j++) {
	line=xpm[1+nColors+j];
	print(" ");
	for(i=0; i < width; i++) {
	  // The following is the index in the color table
	  // corresponding to the character at this position
	    print("%c",line[i]);
	}
	print("\n");
    }
    print("\n");
    for(j=0; j < height; j++) {
	for(i=0; i < width; i++) {
	    print(" %2d",colorData[i+width*j]);
	}
	print("\n");
    }
#endif
    
    return XPM_OK;
}

// Parses a color string to get the colors
int CXpm::parseColor(char *color, int nColor, UChar *red, UChar *green,
  UChar *blue)
{
    int len;
    int offset;
    char colorNum[4];     // FFF
    int status;
    
    if(!color) {
	errMsg("CXpm::parseColor: Color spec is NULL\n");
	return XPM_ERROR;
    }

  // Check for transparent color (Use only the last one)
    if(!strcmp(color,"none") || !strcmp(color,"NONE") ||
      !strcmp(color,"None")) {
	transparentColorFlag=True;
	transparentColorIndex=nColor;
	*red=*green=*blue=0;
	return XPM_OK;
    }

  // Check if a hex specification
  // Could just use XQueryColor if it weren't for exceed 5
    if(*color == '#') {
      // Check if length is a multiple of 3 and is 3 or less
	len=strlen(color+1);
	offset=len/3;
	if(offset*3 != len || offset > 3) {
	    errMsg("CXpm::parseColor: Bad color spec: %s\n",color);
	    return XPM_ERROR;
	}

      // Extract the colors
	colorNum[offset]='\0';
	strncpy(colorNum,color+1,offset);
#if DEBUG_XPM > 1
	print("CXpm::parseColor: color=%s offset=%d colorNum=%s"
	  " atoi(colorNum)=%d\n",
	  color,offset,colorNum,atoi(colorNum));
#endif
	*red=(unsigned char)strtol(colorNum,NULL,16);
	strncpy(colorNum,color+1+offset,offset);
	*green=(unsigned char)strtol(colorNum,NULL,16);
	strncpy(colorNum,color+1+offset*2,offset);
	*blue=(unsigned char)strtol(colorNum,NULL,16);

	return XPM_OK;
    }

  // Check if the color is in the database
    XColor exactDef,screenDef;
    Colormap colormap=DefaultColormap(display,screen);
    
    status=XLookupColor(display,colormap,color,&exactDef,&screenDef);
    if(status) {
	*red=screenDef.red>>8;
	*green=screenDef.green>>8;
	*blue=screenDef.blue>>8;
	return XPM_OK;
    } else {
      // Not found
	errMsg("CXpm::parseColor: Cannot find color %s\n",color);
	*red=*green=*blue=0;
	return XPM_ERROR;
    }
}

// Allocates X colors
int CXpm::allocateColors(void)
{
    Pixel blackPixel=BlackPixel(display,screen);
    Colormap colormap=DefaultColormap(display,screen);
    XColor xColors[256];
    int unallocated=0;

    for(int i=0; i < nColors; i++) {
      // Check for transparent color */
	xColors[i].red=Red[i]<<8;
	xColors[i].green=Green[i]<<8;
	xColors[i].blue=Blue[i]<<8;
	xColors[i].flags=DoRed|DoGreen|DoBlue;
	if(!XAllocColor(display,colormap,&xColors[i])) { 
	    unallocated++;
	    xColors[i].pixel=blackPixel;
	}
	pixels[i]=xColors[i].pixel;
    }
    if(unallocated) {
	errMsg("CXpm: allocateColors: Could not allocate %d of %d colors",
	  unallocated,nColors);
    }

    return XPM_OK;
}

// Writes a color to the color data corresponding to an index in the
// color table, keeping track of the rows and columns
void CXpm::addToColorData(UChar index)
{
    int offset=YC*width+XC;
	
#if DEBUG_COLOR > 1
	print(" index=%3d %02x %02x %02x XC=%2d YC=%2d offset=%3d\n",
	  index,
	  getRed(index),getGreen(index),
	  getBlue(index),XC,YC,offset);
#endif
  // Check that we are in bounds and write to the XImage data
    if(offset < colorDataSize) {
	UChar *p=colorData+offset;
	*p=SAFE(index);
#if DEBUG_COLOR > 1
	print("+index=%3d %02x %02x %02x XC=%2d YC=%2d offset=%3d\n",
	  *p,
	  getRed(*p),getGreen(*p),
	  getBlue(*p),XC,YC,offset);
#endif
    }
	
  // Update the X-coordinate, and if it overflows, update the Y-coordinate
    XC++;
    if(XC >= width) {
	XC=0;
	YC++;
    }
}

// Gets the byte order for this client (No longer used to set the
// ClientByteOrder, which is always MSBFirst)
int CXpm::getClientByteOrder()
{
    short i=1;
    
    return (*(char*)&i == 1) ? LSBFirst : MSBFirst;
}

// Writes a color to the image data corresponding to an index in the
// color table, keeping track of the rows and columns
void CXpm::addToImageData(UChar index)
{
    Pixel clr;
    int offset=YC*bytesPerScanline+XC;
	
  //Check that we are in bounds and write to the XImage data
    if(offset < imageDataSize) {
	UChar *p=(UChar *)imageData+offset;
	
      //Check for transparent color
	if(transparentColorFlag && SAFE(index) ==
	  transparentColorIndex) {
	    clr=backgroundColor;
	} else {
	    clr=pixels[SAFE(index)];
	}
#if DEBUG_COLOR
	print(" index=%3d %02x %02x %02x XC=%2d YC=%2d trans=%s pixel=%x\n",
	  index,
	  getRed(index),getGreen(index),
	  getBlue(index),XC,YC,
	  (transparentColorFlag && SAFE(index) ==
	    transparentColorIndex)?"True":"False",
	  clr);
#endif

      // Write the appropriate part of the pixel in MSBFirst order
	switch(bytesPerPixel) {
	case 4:
	    *p++=(UChar)((clr & 0xff000000) >> 24);
	  /* Fall through */
	case 3:
	    *p++=(UChar)((clr & 0x00ff0000) >> 16);
	  /* Fall through */
	case 2:
	    *p++=(UChar)((clr & 0x0000ff00) >> 8);
	  /* Fall through */
	case 1:
	    *p=(UChar)((clr & 0x000000ff));
	    break;
	}
    }

  //Update the X-coordinate, and if it overflows, update the Y-coordinate
    XC=XC+bytesPerPixel;
    if(XC+bytesPerPixel > bytesPerScanline) {
	XC=0;
	YC++;
    }
}

// Writes the color array data to the image data
void CXpm::writeImageData(void) {
    UChar *iptr=colorData;
    UChar *eof=colorData+colorDataSize;

#if DEBUG_COLOR > 1
    print("CImage::writeImageData: colorDataSize=%d imageDataSize=%d\n",
      colorDataSize,imageDataSize);
#endif

  // Just copy the color indices from the colorArray to the imagedata
    XC=YC=0;
    while(iptr < eof) {
#if DEBUG_COLOR > 1
	print(" index=%3d %02x %02x %02x\n",
	  *iptr,
	  getRed(*iptr),getGreen(*iptr),getBlue(*iptr));
#endif
	addToImageData(*iptr);
	iptr++;
    }
}

int CXpm::redraw(void)
{
  // Free any existing pixmap and make a new one
    if(pixmap) {
	XFreePixmap(display,pixmap);
	pixmap=(Pixmap)NULL;
    }
    pixmap=XCreatePixmap(display,RootWindow(display,screen),width,height,
      screenDepth);
    if(!pixmap) {
	errMsg("CXpm::redraw: Cannot create pixmap");
	return XPM_ERROR;
    }
    
  // Fill pixmap with the background color
    XSetForeground(display,gc,backgroundColor);
    XFillRectangle(display,pixmap,gc,0,0,width,height);	
    
  // Copy the image to the pixmap
    XPutImage(display,pixmap,gc,image,0,0,0,0,width,height);	

    return XPM_OK;
}

void CXpm::printSummary(void) const
{
    print("\nXPM Image:\n"
      "  width=%d height=%d nColors=%d cpp=%d xhot=%d yhot=%d XPMEXT=%s\n",
      width,height,nColors,cpp,xhot,yhot,
      XPMEXT?XPMEXT:"NULL");
    print("  transparentColorFlag=%s transarentColorIndex=%d\n",
      transparentColorFlag?"True":"False",transparentColorIndex);
    print("  colorData=%x imageData=%x image=%x pixmap=%x\n",
      colorData,imageData,image,pixmap);
    print("  backgroundColor=%d bytesPerScanline=%d bytesPerPixel=%d\n",
      backgroundColor,bytesPerScanline,bytesPerPixel);
    print("  screenDepth=%d XBitsPerPixel=%d XBitmapBitOrder=%s\n",
      screenDepth,XBitsPerPixel,
      XBitmapBitOrder==LSBFirst?"LSBFirst":"MSBFirst");
    print("  XBitmapPad=%d XBitmapUnit=%d XByteOrder=%s\n",
      XBitmapPad,XBitmapUnit,
      XByteOrder==LSBFirst?"LSBFirst":"MSBFirst");
    print("  clientByteOrder=%s\n",
      clientByteOrder==LSBFirst?"LSBFirst":"MSBFirst");
    for(int i=0; i < nColors; i++) {
	int r=Red[i],g=Green[i],b=Blue[i];
	print("  Color: %3d Char: \"%c\" Color: %-7s"
	  " Red=%02x Green=%02x Blue=%02x Pixel=%*x\n",
	  i,colorChar[i],
	  colorDef[i]?colorDef[i]:"NULL",r,g,b,
	  screenDepth/4,pixels[i]);
    }
}

int CXpm::errMsgDefault(const char *fmt, ...)
{
    static char string[2048];
    va_list vargs;
    int nchars=0;

    va_start(vargs,fmt);
    nchars+=vsprintf(string,fmt,vargs);
    va_end(vargs);

#ifdef WIN32
    print("%s\n",string);
#else
    fprintf(stderr,"%s\n",string);
#endif

    return nchars;
}


