/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// Header file for XPM

// CXpm is a class to handle XPM formats.  It is here because the XPM
// routines are not part of the X distribution, and the XPM library
// and header file is often not available.  This class does not
// purport to support the entire XPM capability, but should handle
// your own files (assumed to be small).  If not it may be easier to
// fix the files rather than install the true XPM routines.

// Restrictions:
//   Will only handle 1 pixel per color.
//   Will only handle 256 colors or fewer
//   Ignores color keywords (m,c,g4,g,s).
//   Ignores symbolic names

// The following must be 1024 or less for WIN32
#define FIXED_MSG_SIZE 1024
#define MAX_COLOR_STRING_LEN 14   // #0000ffff0000  32-bits per pixel

#define XPM_OK 0
#define XPM_ERROR 1;

#ifdef WIN32
// Hummingbird extra functions including lprintf
//   Needs to be included after Intrinsic.h for Exceed 5
# include <stdio.h>
# include <X11/Intrinsic.h>
# include <X11/XlibXtra.h>
// The following is done in Exceed 6 but not in Exceed 5
//   Need it to define printf as lprintf for Windows
//   (as opposed to Console) apps */
# ifdef _WINDOWS
#  ifndef printf
#   define printf lprintf
#  endif    
# endif    
#else // #ifdef WIN32
// WIN32 does not have unistd.h
# include <unistd.h>
# include <limits.h>
#endif // #ifdef WIN32

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <Xm/Xm.h>

typedef unsigned char UChar;

typedef int (*XpmErrMsg)(const char *fmt, ...);

/**************************************************************************/
/**************************** CXpm ****************************************/
/**************************************************************************/

class CXpm
{
  public:
    CXpm(Display *displayIn, char **xpmIn, Pixel bg=0,
      XpmErrMsg errMsgIn=errMsgDefault);
    ~CXpm(void);
    int getNColors(void) const { return nColors; }
    int getBackgroundColor(void) const { return backgroundColor; }
    UChar getRed(int i) const { return Red[i]; }
    UChar getGreen(int i) const { return Green[i]; }
    UChar getBlue(int i) const { return Blue[i]; }
    const Pixel *getPixels(void) const { return pixels; }
    Pixmap getPixmap(void) const { return pixmap; }
    Visual *getVisual(void) const { return visual; }
    int getClientByteOrder();
    void printSummary(void) const;

  private:
    int parse(void);
    int parseColor(char *color, int nColor, UChar *red, UChar *green,
      UChar *blue);
    int allocateColors(void);
    void addToColorData(UChar Index);
    void addToImageData(UChar Index);
    void writeImageData(void);
    int redraw(void);

    static int errMsgDefault(const char *fmt, ...);

  private:
    XpmErrMsg errMsg;
    char **xpm;
    int width;
    int height;
    int nColors;
    int cpp;
    int xhot;
    int yhot;
    char *XPMEXT;

    char colorDef[256][MAX_COLOR_STRING_LEN];
    UChar colorIndex[256];           // Index in color table for character
    UChar colorChar[256];            // Character for index in the color table
    UChar Red[256], Green[256], Blue[256];  // RGB values for the color table
    Boolean transparentColorFlag;     // Flag for transparent color
    UChar transparentColorIndex;      // Index of the transparent color
    Pixel pixels[256];                // Pixels for the color table

    Pixmap pixmap;                    // Pixmap for drawing area repaints
    UChar *imageData;                 // Data for the X image
    UChar *colorData;                 // Data for the color array
    int imageDataSize;                // The size of the XImage data
    int colorDataSize;                // The size of the color data
    XImage *image;                    // The X image
    GC gc;                            // Graphics context for drawing area
    int XC, YC;                       // Output X and Y coords of current pixel
    int backgroundColor;              // background color

  // X Parameters (Order is significant for initialization)
    Display *display;                 // X display
    int screen;                       // X default screen
    int screenDepth;                  // X Screen depth
    int XBitsPerPixel;                // X bits per pixel
    int XBitmapBitOrder;              // XImage bitmap bit order
    int XBitmapPad;                   // XImage bitmap pad
    int XBitmapUnit;                  // XImage bitmap unit
    int XByteOrder;                   // XImage byte order
    Visual *visual;                   // Visual

  // Image parameters
    int bytesPerScanline;             // Bytes per scanline in output
    int bytesPerPixel;                // Bytes per pixel

  // Machine parameters
    int clientByteOrder;
};
